#region Using

using System;
using System.IO;
using System.Web;
using System.Net;
using System.Xml;
using System.Collections.Generic;

#endregion

/// <summary>
/// Summary description for FoafParser
/// </summary>
public class FoafParser
{

	#region Properties

	private List<Person> _Friends = new List<Person>();
	/// <summary>
	/// Gets or sets the friends.
	/// </summary>
	/// <value>The friends.</value>
	public List<Person> Friends
	{
		get { return _Friends; }
	}

	private Person _Owner;

	public Person Owner
	{
		get { return _Owner; }
	}

	private bool _IsParsed;
	public bool IsParsed
	{
		get { return _IsParsed; }
	}

	#endregion

	/// <summary>
	/// Parses the FOAF content found at the specified address.
	/// </summary>
	/// <param name="address">The address of the FOAF document.</param>
	public void Parse(Uri address)
	{
		XmlDocument doc = RetrieveXml(address);

		if (doc != null)
		{
			XmlNamespaceManager nsMgr = GetNamespaceManager(doc);
			XmlNode personNode = doc.SelectSingleNode("//foaf:Person", nsMgr);
			_Owner = ConstructPersonFromFoaf(personNode, address, null);

			XmlNodeList knows = doc.SelectNodes("//foaf:knows/foaf:Person", nsMgr);

			foreach (XmlNode node in knows)
			{
				string location = ParseNode(node.SelectSingleNode("s:seeAlso", nsMgr));
				Uri url;
				Uri.TryCreate(location, UriKind.Absolute, out url);

				Person person = ConstructPersonFromFoaf(node, url, null);
				if (person != null)
					Friends.Add(person);
			}

			_IsParsed = true;
		}
	}

	/// <summary>
	/// Constructs the person from an XML node in the FOAF file.
	/// </summary>
	private Person ConstructPersonFromFoaf(XmlNode node, Uri foafLocation, Person person)
	{
		XmlNamespaceManager nsMgr = GetNamespaceManager(node.OwnerDocument);

		if (node != null)
		{
			if (person == null)
				person = new Person();

			person.FoafLocation = foafLocation;
			person.Title = ParseNode(node.SelectSingleNode("foaf:title", nsMgr));
			person.Name = ParseNode(node.SelectSingleNode("foaf:name", nsMgr));
			person.FirstName = ParseNode(node.SelectSingleNode("foaf:givenname", nsMgr));
			person.FamilyName = ParseNode(node.SelectSingleNode("foaf:family_name", nsMgr));
			person.NickName = ParseNode(node.SelectSingleNode("foaf:nick", nsMgr));

			person.OpenID = ParseNode(node.SelectSingleNode("foaf:openid", nsMgr));
			person.EmailAddress = ParseNode(node.SelectSingleNode("foaf:mbox", nsMgr));
			person.EmailChecksum = ParseNode(node.SelectSingleNode("foaf:mbox_sha1sum", nsMgr));

			person.Weblog = ParseNode(node.SelectSingleNode("foaf:weblog", nsMgr));
			person.Homepage = ParseNode(node.SelectSingleNode("foaf:homepage", nsMgr));
			person.Gender = ParseNode(node.SelectSingleNode("foaf:gender", nsMgr));
			person.Photo = ParseNode(node.SelectSingleNode("foaf:depiction", nsMgr));

			if (string.IsNullOrEmpty(person.Photo))
				person.Photo = ParseNode(node.SelectSingleNode("foaf:img", nsMgr));

			if (string.IsNullOrEmpty(person.Photo))
				person.Photo = ParseNode(node.SelectSingleNode("foaf:image", nsMgr));

			DateTime birthday;
			DateTime.TryParse(ParseNode(node.SelectSingleNode("foaf:dateOfBirth", nsMgr)), out birthday);
			person.Birthday = birthday;

			return person;
		}

		return null;
	}

	#region Parse friends async

	private int _Counter;
	private int _NumberOfFriendsWithFoaf;

	public void ParseFriendsAsync()
	{
		List<Person> friends = Friends.FindAll(delegate(Person p) { return p.FoafLocation != null; });
		_NumberOfFriendsWithFoaf = friends.Count;
		_Counter = 0;

		friends.ForEach(delegate(Person p)
		{
			BeginRetrieveFriend(p);
		});
	}

	private class AsyncTransport
	{
		public WebRequest Request;
		public Person Friend;
		public FoafParser Parser;
	}

	private void BeginRetrieveFriend(Person friend)
	{
		WebRequest request = WebRequest.Create(friend.FoafLocation);
		request.Timeout = 5000;

		AsyncTransport data = new AsyncTransport();
		data.Request = request;
		data.Friend = friend;
		data.Parser = this;

		request.BeginGetResponse(EndRetrieveFriend, data);
	}

	private void EndRetrieveFriend(IAsyncResult result)
	{
		AsyncTransport data = (AsyncTransport)result.AsyncState;
		try
		{
			using (WebResponse response = data.Request.GetResponse())
			using (Stream stream = response.GetResponseStream())
			{
				XmlDocument doc = new XmlDocument();
				doc.Load(stream);

				XmlNamespaceManager nsMgr = GetNamespaceManager(doc);
				XmlNode personNode = doc.SelectSingleNode("//foaf:Person", nsMgr);

				ConstructPersonFromFoaf(personNode, data.Request.RequestUri, data.Friend);
			}
		}
		catch (Exception)
		{
		}
		finally
		{
			OnFriendParsed(data.Friend);
			data.Parser._Counter++;

			if (data.Parser._NumberOfFriendsWithFoaf == data.Parser._Counter)
				OnComplete();
		}
	}

	#endregion

	#region XML helper methods

	/// <summary>
	/// Retrieves the XML from a web location and loads it into an XmlDocument.
	/// </summary>
	/// <param name="address">The address.</param>
	/// <returns></returns>
	private XmlDocument RetrieveXml(Uri address)
	{
		try
		{
			using (WebClient client = new WebClient())
			using (Stream stream = client.OpenRead(address))
			{
				XmlDocument doc = new XmlDocument();
				doc.Load(stream);
				return doc;
			}
		}
		catch (Exception)
		{
			return null;
		}
	}

	/// <summary>
	/// Gets the namespace manager with the supported namespaces.
	/// </summary>
	private static XmlNamespaceManager GetNamespaceManager(XmlDocument doc)
	{
		XmlNamespaceManager nsMgr = new XmlNamespaceManager(doc.NameTable);

		nsMgr.AddNamespace(string.Empty, "http://xmlns.com/foaf/0.1/");
		nsMgr.AddNamespace("foaf", "http://xmlns.com/foaf/0.1/");
		nsMgr.AddNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
		nsMgr.AddNamespace("s", "http://www.w3.org/2000/01/rdf-schema#");
		//nsMgr.AddNamespace("cc", "http://creativecommons.org/ns#");
		//nsMgr.AddNamespace("con", "http://www.w3.org/2000/10/swap/pim/contact#");
		//nsMgr.AddNamespace("dc", "http://purl.org/dc/elements/1.1/");
		//nsMgr.AddNamespace("geo", "http://www.w3.org/2003/01/geo/wgs84_pos#");
		//nsMgr.AddNamespace("owl", "http://www.w3.org/2002/07/owl#");		

		return nsMgr;
	}

	/// <summary>
	/// Parses the node for foaf and rdfs content.
	/// </summary>
	/// <param name="node">The node containing the value.</param>
	/// <returns>A string value of the contents of the node.</returns>
	private string ParseNode(XmlNode node)
	{
		if (node != null)
		{
			if (!string.IsNullOrEmpty(node.InnerText))
			{
				return node.InnerText;
			}
			else if (node.Attributes["rdf:resource"] != null)
			{
				return node.Attributes["rdf:resource"].InnerText;
			}
		}

		return null;
	}

	#endregion

	#region Events

	/// <summary>
	/// Occurs when the parser has retrieved all information from the FOAF file.
	/// </summary>
	public event EventHandler<EventArgs> FriendParsingComplete;
	protected void OnComplete()
	{
		if (FriendParsingComplete != null)
		{
			FriendParsingComplete(this, new EventArgs());
		}
	}

	/// <summary>
	/// Occurs when a friend has been parsed and added to the Friend collection.
	/// </summary>
	public event EventHandler<EventArgs> FriendParsed;
	protected void OnFriendParsed(Person person)
	{
		if (FriendParsed != null)
		{
			FriendParsed(person, new EventArgs());
		}
	}

	#endregion

}

/// <summary>
/// A simplified person class. 
/// You might want to expand or replace it with your own.
/// </summary>
public class Person
{
	public string Title;
	public string Name;
	public string FirstName;
	public string FamilyName;
	public string NickName;

	public string Weblog;
	public string Homepage;

	public string Photo;
	public string OpenID;
	public string Gender;
	public DateTime Birthday;
	public Uri FoafLocation;

	public string EmailAddress;
	public string EmailChecksum;
}